from .amatrix import makeA, makeD
from .gmatrix import makeG
from .hmatrix import makeH
from .sort import sortPed
from .loaddata import loadEgPed, loadEgGeno
from .selectped import selectPed
from .plot import cluster, heat,gragh, pca
from .coef import coefKinship,coefInbreeding
from .inter_matrix import makeG_inter
from .mmatrix import makeM